# -*- coding: utf-8 -*-
"""
Module for manipulating the Sentinel-2 tiles index geopackage file.
"""

import logging
import pickle
import fiona
from osgeo import ogr
from shapely.geometry import shape, Point, box
# type annotation
from typing import List, Set, Dict, Optional
from pathlib import Path


from .config import SHARED_DATA, Config
from .library import Library
from .tiles import Tile


logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

TILES_INDEX = SHARED_DATA.get("tiles_index")
TILES_INDEX_DICT = SHARED_DATA.get("tiles_index_dict")


def serialise_tiles_index() -> None:
    """Saves sentinel_2_index.gpkg as a pickle file if it's missing."""
    tiles_dict = dict()

    with fiona.open(str(TILES_INDEX), "r") as tiles_index:
        for tile in tiles_index:
            name = tile["properties"]["Name"]
            geom = shape(tile['geometry'])
            tiles_dict[name] = geom

    with open(str(TILES_INDEX_DICT), "wb") as pfile:
        pickle.dump(tiles_dict, pfile)


def get_tiles_from_point(lon: float,
                         lat: float,
                         land_only: bool = False
                         ) -> Set[str]:
    """
    Returns tiles containing a point (longitude, latitude)

    :param lon: longitude.
    :param lat: latitude.
    :param land_only: keep only tiles that contain lands.
    """
    intersected_tiles = set()

    with fiona.open(str(TILES_INDEX), 'r') as tiles_index:
        point_geom = Point(lon, lat)
        if land_only:
            tiles_index = filter(lambda feat: feat["properties"]["land"] == "True", tiles_index)

        for tile in tiles_index:
            tile_geom = shape(tile['geometry'])
            if tile_geom.contains(point_geom):
                intersected_tiles.add(tile["properties"]["Name"])
    return intersected_tiles


def get_tiles_from_bbox(lon_min: float, lat_min: float,
                        lon_max: float, lat_max: float,
                        land_only=False
                        ) -> Set[str]:
    """
    Returns S2 tiles intersecting a bbox.

    :param lon_min: longitude.
    :param lat_min: latitude.
    :param lon_max: longitude.
    :param lat_max: latitude.
    :param land_only: keep only tiles that contain lands.
    """
    intersected_tiles = set()

    with fiona.open(str(TILES_INDEX), 'r') as tiles_index:
        bbox_geom = box(lon_min, lat_min, lon_max, lat_max)
        if land_only:
            tiles_index = filter(lambda feat: feat["properties"]["land"] == "True", tiles_index)

        for tile in tiles_index:
            tile_geom = shape(tile['geometry'])
            if tile_geom.intersects(bbox_geom):
                intersected_tiles.add(tile["properties"]["Name"])

    return intersected_tiles


def get_tiles_from_file(
        vectors_file: str,
        land_only: bool = False
) -> Dict[str, Optional[List[str]]]:
    """
    Returns the intersected tiles for each feature id in the vectors file.

    :param vectors_file: path to a file containing vectors geometries.
    :param land_only: return only tiles that contain lands.
    """
    # OGR is used for better performances
    drv_gpkg = ogr.GetDriverByName("GPKG")
    tiles_index_ds = drv_gpkg.Open(str(TILES_INDEX), 0)
    tiles_index_lyr = tiles_index_ds.GetLayer(0)
    vectors_file_ds = ogr.Open(str(vectors_file), 0)
    vectors_file_lyr = vectors_file_ds.GetLayer(0)

    # filter on tiles with land
    if land_only:
        tiles_index_lyr.SetAttributeFilter("land = 'True'")

    # this dict will contain the geometry id as key and the list of
    # tiles it intersects as value
    intersected_tiles_dict = dict()  # type: Dict[str, Optional[List[str]]]
    feat_id = 0
    for feat in vectors_file_lyr:
        feat_geom = feat.GetGeometryRef()
        tiles_index_lyr.SetSpatialFilter(feat_geom)

        contained_in_tiles = []
        for filtered_t in tiles_index_lyr:
            tile = filtered_t.GetField("Name")
            contained_in_tiles.append(tile)

        intersected_tiles_dict[str(feat_id)] = contained_in_tiles if contained_in_tiles else None
        feat_id += 1

    return intersected_tiles_dict


def get_processed_indices_vect(
        out_shapefile: str = None
):
    """
    Returns a vector file of the indice processed tiles

    :param out_shapefile: path to a output file vector.
    """
    if not out_shapefile:
        out_temp_path = Path(Config().get("temp_path"))
        out_shapefile = str(out_temp_path / "tile_indices_shp.shp")
    drv_gpkg = ogr.GetDriverByName("GPKG")
    input_layer_ds = drv_gpkg.Open(str(TILES_INDEX), 0)
    input_layer_lyr = input_layer_ds.GetLayer(0)
    
    ndvi_index = Library().ndvi
    ndwigao_index = Library().ndwigao
    ndwimcf_index = Library().ndwimcf
    mndwi_index = Library().mndwi
    indices_index = set().union(ndvi_index, ndwigao_index, ndwimcf_index, mndwi_index)
    
    tuile_count = {}
    for key in indices_index:
        tuile_count[key]={'ndvi': 0, 'ndwigao': 0, 'ndwimcf': 0, 'mndwi': 0} 
    
    for key in ndvi_index:
        tuile_count[key]['ndvi'] = len(Tile(key).ndvi.raws)
    for key in ndwigao_index:
        tuile_count[key]['ndwigao'] = len(Tile(key).ndwigao.raws)
    for key in ndwimcf_index:
        tuile_count[key]['ndwimcf'] = len(Tile(key).ndwimcf.raws)
    for key in mndwi_index:
        tuile_count[key]['mndwi'] = len(Tile(key).mndwi.raws)
    
    # pour supprimer les tuiles ayant tous les indices à 0
    tuile_count = {k: v for k, v in tuile_count.items() if sum(v[indice] for indice in v.keys())}

    query_str = 'or '.join(['"{}" = "{}"'.format("Name", idx) for idx in tuile_count.keys()])
    input_layer_lyr.SetAttributeFilter(query_str)

    #~ input_layer_lyr.CreateField(field_name)
    
    driver = ogr.GetDriverByName('ESRI Shapefile')
    out_ds = driver.CreateDataSource(out_shapefile)
    out_layer = out_ds.CopyLayer(input_layer_lyr, ndvi_index[0])
    
    for name in ["NDVI", "NDWIGAO", "NDWIMCF", "MNDWI"]:
        field_name = ogr.FieldDefn(name, ogr.OFTString)
        field_name.SetWidth(10)
        out_layer.CreateField(field_name)
    
    for feat in out_layer:
        tile = feat.GetField("Name")
        #~ logger.info(tile)
        feat.SetField("NDVI", tuile_count[tile]['ndvi'])
        feat.SetField("NDWIGAO", tuile_count[tile]['ndwigao'])
        feat.SetField("NDWIMCF", tuile_count[tile]['ndwimcf'])
        feat.SetField("MNDWI", tuile_count[tile]['mndwi'])
        out_layer.SetFeature(feat)
    
    out_layer = None
    del input_layer_ds, input_layer_lyr, out_layer, out_ds
    
