# -*- coding: utf-8 -*-

"""
Module for managing products and tiles in the library and temp folders.
"""

import os
import shutil
import logging
import pathlib

from pathlib import Path
# type annotations
from typing import List, Dict

from .config import Config, SHARED_DATA
from .products import L1cProduct
from .tiles import Tile


s2_tiles_index = SHARED_DATA.get("tiles_index")

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class Library:
    """Class for listing L1C, L2A and indices tiles in the library folders.

    """
    _temp_path = Path(Config().get("temp_path"))
    _l1c_path = Path(Config().get("l1c_path"))
    _l2a_path = Path(Config().get("l2a_path"))
    _indices_path = Path(Config().get("indices_path"))

    def __init__(self):
        self._l1c = [f.name for f in self._l1c_path.glob("*")]
        self._l2a = [f.name for f in self._l2a_path.glob("*")]
        self._indices = dict()

        for indice in [f.name for f in self._indices_path.glob("*")]:
            self.__dict__[indice.lower()] = [f.name for f in self._indices_path.glob("{}/*".format(indice))]

    @property
    def l1c(self) -> List[str]:
        """Returns tiles in the L1C library folder."""
        return self._l1c

    @property
    def l2a(self) -> List[str]:
        """Returns tiles in the L2A library folder."""
        return self._l2a

    @property
    def indices(self) -> Dict[str, pathlib.PosixPath]:
        """Returns tiles in the indices library folder."""
        return self._indices
        
    def clean(self, 
              clean_list : list = [], 
              remove: bool = False):
        if not clean_list:
            clean_list = self.l1c
        for t in clean_list:
            try:
                til = Tile(t)
                til.clean_lib(remove=remove)
            except:
                pass


class TempContainer:
    """Class for managing a downloaded L1C products.

    :param l1c_identifier: L1C product's identifier.
    :param tile: product's tile name.
    """
    _temp_path = Path(Config().get("temp_path"))

    def __init__(self, l1c_identifier=None, tile: str = None):
        self.tile = tile
        self.temp_id = l1c_identifier + "_TEMP"
        self.temp_path = self._temp_path / self.temp_id
        self.l1c = L1cProduct(identifier=l1c_identifier, tile=tile, path=self.temp_path)

    def create_temp_folder(self) -> "TempContainer":
        """Creates the L1C product's temporary folder in the TEMP library folder."""
        logger.info("{}: creating temp folder".format(self.l1c.identifier))
        try:
            self.temp_path.mkdir(exist_ok=True)
        except FileNotFoundError:
            print("Invalid temp folder path:", self.temp_path)
            raise
        return self

    def unzip_l1c(self) -> "TempContainer":
        """Unzips the downloaded L1C zip file in the product's temporary folder."""
        zip_file = self.temp_path / (self.l1c.identifier + ".zip")
        if zip_file.exists():
            logger.info("{} : Unzipping L1C archive".format(self.l1c.identifier))
            shutil.unpack_archive(str(zip_file), str(self.temp_path))
            os.remove(str(zip_file))
        return self

    def delete_temp_folder(self) -> "TempContainer":
        """Deletes the L1C product's temporary folder."""
        logger.info("Deleting temp folder: {}".format(self.temp_path))
        if not list(self.temp_path.glob("*.SAFE")):
            try:
                shutil.rmtree(str(self.temp_path))
            except FileNotFoundError:
                logger.warning("Temp folder does not exist: {}".format(self.temp_path))
        else:
            logger.error("Temp folder contains a SAFE product.")
        return self
