# -*- coding: utf-8 -*-

"""
Module for managing products and tiles in the library and temp folders.
"""
import pathlib
import logging
import re
import fiona
import shutil

from pathlib import Path
from collections import namedtuple
from datetime import datetime
from pprint import pformat
# type annotations
from typing import List, Dict, Iterable

from .config import Config, SHARED_DATA
from .utils import str_to_datetime
from .indices import IndicesCollection
from .products import L1cProduct, L2aProduct


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


# namedtuple
TileProduct = namedtuple("TileProduct", ["identifier", "date", "cloud_cover"])


class ProductsList:
    """Class for managing tiles products lists. Each item is a TileProduct
    namedtuple.

    """
    def __init__(self):
        self._dict = dict()
        self._time_index = dict()

    @property
    def products(self) -> List[str]:
        """Returns products names."""
        return list(self._dict.keys())

    @property
    def dates(self) -> List[datetime]:
        """Returns products dates."""
        return list(self._time_index.keys())

    @property
    def first(self) -> TileProduct:
        """Returns oldest product."""
        try:
            min_date = min(self.dates)
        except ValueError:
            return TileProduct(None, None, None)
        prod = self._time_index[min_date]
        return TileProduct(prod, min_date, self._dict[prod]["cloud_cover"])

    @property
    def last(self) -> TileProduct:
        """Returns newest product."""
        try:
            max_date = max(self.dates)
        except ValueError:
            return TileProduct(None, None, None)
        prod = self._time_index[max_date]
        return TileProduct(prod, max_date, self._dict[prod]["cloud_cover"])

    def filter_dates(
            self,
            date_min: str = None, date_max: str = None
    ) -> "ProductsList":
        """Filters products list in a time range.

        :param date_min: oldest date.
        :param date_max: newest date.
        """
        min_date = str_to_datetime(date_min, "ymd") if date_min else self.first.date
        max_date = str_to_datetime(date_max, "ymd") if date_max else self.last.date
        filtered = ProductsList()
        for k, v in self._dict.items():
            if min_date.date() <= v["date"].date() <= max_date.date():
                filtered[k] = {"date": v["date"], "cloud_cover": v["cloud_cover"]}
        return filtered

    def filter_clouds(
            self,
            cover_min: int = 0, cover_max: int = 100
    ) -> "ProductsList":
        """Filters products list.

        :param cover_min: minimum cloud coverage.
        :param cover_max: maximum cloud coverage.
        """
        filtered = ProductsList()
        for k, v in self._dict.items():
            if cover_min <= v["cloud_cover"] <= cover_max:
                filtered[k] = {"date": v["date"], "cloud_cover": v["cloud_cover"]}
        return filtered

    def __len__(self) -> int:
        return len(self._dict)

    def __contains__(self, item) -> bool:
        return item in self._dict

    def __repr__(self) -> str:
        return pformat([self[k] for k in self.products], indent=1, width=1)

    def __iter__(self) -> Iterable:
        return (self[k] for k in self._dict)

    def __getitem__(self, item) -> TileProduct:
        if item in self._dict:
            return TileProduct(item, self._dict[item]["date"], self._dict[item]["cloud_cover"])
        else:
            raise KeyError("{} not found".format(item))

    def __setitem__(self, item, value) -> None:
        self._dict[item] = {"date": value["date"], "cloud_cover": value["cloud_cover"]}
        self._time_index[value["date"]] = item


class IndicesList(ProductsList):
    """Class for managing indices products lists.

    """
    @property
    def raws(self) -> "ProductsList":
        filtered = ProductsList()
        for k, v in self._dict.items():
            if not("MASK" in k) and not("QUICKLOOK" in k):
                filtered[k] = {"date": v["date"], "cloud_cover": v["cloud_cover"]}
        return filtered

    @property
    def masks(self) -> "ProductsList":
        filtre = ProductsList()
        for k, v in self._dict.items():
            if "MASK" in k:
                filtre[k] = {"date": v["date"], "cloud_cover": v["cloud_cover"]}
        return filtre

    @property
    def quicklooks(self) -> "ProductsList":
        filtered = ProductsList()
        for k, v in self._dict.items():
            if "QUICKLOOK" in k:
                filtered[k] = {"date": v["date"], "cloud_cover": v["cloud_cover"]}
        return filtered


class Tile:
    """Class for managing tiles in the library.

    :param name: tile's name.

    Usage:
        >>> Tile("40KCC")
    """
    def __init__(self, name: str):
        self.name = name
        self._paths = {"l1c": Path(Config().get("l1c_path")) / name,
                       "l2a": Path(Config().get("l2a_path")) / name,
                       "indices": {}
                       }
        self._indices_path = Path(Config().get("indices_path"))

        self._products = {"l1c": ProductsList(),
                          "l2a": ProductsList(),
                          "cloudmasks" : ProductsList(),
                          "indices": dict()}

        self._get_indices_paths()
        self._get_l1c_list()
        self._get_l2a_list()
        self._get_cloudmasks()
        self._get_indices_list()

    def _get_indices_paths(self) -> None:
        """Updates self._paths"""
        for indice in self._indices_path.glob("*/{}".format(self.name)):
            self._paths["indices"][str(indice.parent.name.lower())] = indice

    def _get_l1c_list(self) -> None:
        """Scans L1C folder and adds products in a ProductsList."""
        for f in self._paths["l1c"].glob("*L1C_*.SAFE"):
            cloud_cover = L1cProduct(identifier=f.name, tile=self.name, path=None).cloud_coverage_assessment
            self._products["l1c"][f.name] = {"date": Tile._get_date(f.name), "cloud_cover": float(cloud_cover)}

    def _get_l2a_list(self) -> None:
        """Scans L2A folder and adds products in a ProductsList."""
        for f in self._paths["l2a"].glob("*L2A*.SAFE"):
            l1c_name = f.name.replace("L2A_", "L1C_").replace("_USER_", "_OPER_")
            try:
                date = self._products["l1c"][l1c_name].date
                cloud_cover = self._products["l1c"][l1c_name].cloud_cover
            except KeyError:
                date = Tile._get_date(f.name)
                cloud_cover = L2aProduct(f.name, self.name).cloud_coverage_assessment
            self._products["l2a"][f.name] = {"date": date, "cloud_cover": float(cloud_cover)}

    def _get_cloudmasks(self) -> None:
        """Scans L2A folder for cloud masks and adds corresponding L2A products in a ProductsList."""
        for f in self._paths["l2a"].glob("*L2A*_CLOUD_MASK.jp2"):
            l1c_name = f.name.replace("L2A_", "L1C_").replace("_USER_", "_OPER_").replace("_CLOUD_MASK.jp2", ".SAFE")
            try:
                date = self._products["l1c"][l1c_name].date
                cloud_cover = self._products["l1c"][l1c_name].cloud_cover
            except KeyError:
                date = Tile._get_date(f.name.replace("_CLOUD_MASK.jp2", ".SAFE"))
                cloud_cover = L2aProduct(f.name.replace("_CLOUD_MASK.jp2", ".SAFE"), self.name).cloud_coverage_assessment
            self._products["cloudmasks"][f.name.replace("_CLOUD_MASK.jp2", ".SAFE")] = {"date": date, "cloud_cover": float(cloud_cover)}

    def _get_indices_list(self) -> None:
        """Scans indices folders and adds products in a IndicesList."""
        for indice, path in self._paths["indices"].items():
            if path.is_dir():
                self._products["indices"][indice] = IndicesList()
                indice_template = IndicesCollection.get_indice_cls(indice.upper()).filename_template
                indice_ext = IndicesCollection.get_indice_cls(indice.upper()).ext
                file_patterns = [indice_ext, 'QUICKLOOK.tif']
                files_selected = []
                for p in file_patterns:
                    files_selected.extend(path.glob("*/*{}".format(p)))
                for f in files_selected:
                    try:
                        indice_pattern = re.sub("{.*?}", "", indice_template)
                        remove_pattern = "{}.*".format(indice_pattern)
                        l2a_name = re.sub(remove_pattern, '', f.name) + ".SAFE"
                        date = self._products["l2a"][l2a_name].date
                        cloud_cover = self._products["l2a"][l2a_name].cloud_cover
                    except KeyError:
                        date = Tile._get_date(f.name)
                        cloud_cover = None
                    self._products["indices"][indice][f.name] = {"date": date, "cloud_cover": cloud_cover}
                self.__dict__[indice] = self._products["indices"][indice]

    @staticmethod
    def is_tile_name_valid(tile_name: str) -> bool:
        """Checks if the tile_name is a valid Sentinel-2 tile.

        :param tile_name: tile's name.
        """
        if not re.match(r"^[0-9]{2}[A-Z]{3}$", tile_name):
            return False
        with fiona.open(str(SHARED_DATA["tiles_index"]), 'r') as tiles_index:
            tiles_names = [tile["properties"]["Name"] for tile in tiles_index]
            return tile_name in tiles_names

    @staticmethod
    def _get_date(product_name: str) -> datetime:
        """Extracts the acquisition date of a products filename.
        """
        date_pattern = re.findall(r"[0-9]{8}T[0-9]{6}", product_name)
        if product_name.startswith("S2A_OPER"):
            date = date_pattern[1]
        else:
            date = date_pattern[0]
        return str_to_datetime(date, "filename")

    def __repr__(self) -> str:
        return "Tile('{}')".format(self.name)

    def __str__(self) -> str:
        return self.name

    @property
    def paths(self) -> Dict[str, pathlib.PosixPath]:
        """Returns all the paths related to a Tile object."""
        return self._paths

    @property
    def l1c(self) -> "ProductsList":
        """Returns tile's L1C products as a ProductsList."""
        return self._products["l1c"]

    @property
    def l2a(self) -> "ProductsList":
        """Returns tile's L2A products as a ProductsList."""
        return self._products["l2a"]

    @property
    def cloudmasks(self) -> "ProductsList":
        """Returns tile's cloud masks products as a ProductsList."""
        return self._products["cloudmasks"]

    @property
    def l1c_missings(self) -> "ProductsList":
        """Returns tile's L2A products that don't have a L1C as a ProductsList."""
        prods_list = ProductsList()
        missings_l1c_set = set(self.l2a.products) - {identifier.replace("L1C_", "L2A_").replace("__OPER__", "_USER_")
                                                     for identifier in self.l1c.products}
        for prod in missings_l1c_set:
            prods_list[prod] = {"date": self._products["l2a"][prod].date,
                                "cloud_cover": self._products["l2a"][prod].cloud_cover}
        return prods_list

    @property
    def l2a_missings(self) -> "ProductsList":
        """Returns tile's L1C products that don't have a L2A as a ProductsList."""
        prods_list = ProductsList()
        missings_l2a_set = set(self.l1c.products) - {identifier.replace("L2A_", "L1C_").replace("_USER_", "_OPER_")
                                                     for identifier in self.l2a.products}
        for prod in missings_l2a_set:
            prods_list[prod] = {"date": self._products["l1c"][prod].date,
                                "cloud_cover": self._products["l1c"][prod].cloud_cover}
        return prods_list

    @property
    def cloudmasks_missings(self) -> "ProductsList":
        """Returns tile's L2A products that don't have a cloud mask as a ProductsList."""
        prods_list = ProductsList()
        missings_l2a_set = set(self.l2a.products) - {identifier for identifier in self.cloudmasks.products}
        for prod in missings_l2a_set:
            prods_list[prod] = {"date": self._products["l2a"][prod].date,
                                "cloud_cover": self._products["l2a"][prod].cloud_cover}
        return prods_list

    def missing_indices(self, indice: str) -> "ProductsList":
        """Returns tile's L2A products that don't have indices as a ProductsList."""
        prods_list = ProductsList()
        try:
            missings_indice_set = set(self.l2a.products) - {identifier.replace("_" + indice.upper() + ".jp2", ".SAFE") \
                                                        for identifier in getattr(getattr(self, indice.lower()), 'raws').products}
        except:
            missings_indice_set = set(self.l2a.products)
        for prod in missings_indice_set:
                prods_list[prod] = {"date": self._products["l2a"][prod].date,
                                "cloud_cover": self._products["l2a"][prod].cloud_cover}            
        return prods_list

    def clean_lib(self,
                  remove: bool = False):
        """
        Clean processing errors
        - unmoved error l2a products from l1c folder
        - moved  error l2a products from l2a folder
        - cloud masks error (0kb)
        - indices error (0kb)        
        """
        #~ logger.info("Cleaning {} library".format(self.name))
        # identify residual l2a from l1c folder
        for f in self._paths["l1c"].glob("*L2A*.SAFE"):
            logger.info("Identified {} from L1C folder".format(f.name))
            if remove:
                try:
                    shutil.rmtree(str(f))
                    logger.info("Removing {} from L1C folder".format(f.name))
                except:
                    logger.error("Can't remove {} from L1C folder".format(f.name))
        # identify residual l2a from l1c folder
        for f in self._paths["l2a"].glob("*L2A*.SAFE"):
            if len(list(f.glob("GRANULE/*/IMG_DATA/R10m/*.jp2"))) == 0:
                logger.info("Corrupted L2A {} in L2A folder".format(f.name))
                if remove:
                    try:
                        shutil.rmtree(str(f))
                        logger.info("Removing corrupted L2A {} from L2A folder".format(f.name))
                    except:
                        logger.error("Can't remove {} from L2A folder".format(f.name))
        # identify 0B cloud masks
        for f in self._paths["l2a"].glob("*L2A*_CLOUD_MASK.jp2"):
            if f.stat().st_size == 0:
                logger.info("Corrupted cloud mask {} in L2A folder".format(f.name))
                if remove:
                    logger.info("Removing corrupted cloud mask {} from L2A folder".format(f.name))
                    f.unlink()
        # identify wrong size l2a_QL
        for f in self._paths["l2a"].glob("*L2A*_QL.tif"):
            if f.stat().st_size != 3617212:
                logger.info("Corrupted L2A QL {} in L2A folder".format(f.name))
                if remove:
                    logger.info("Removing corrupted QL {} from L2A folder".format(f.name))
                    f.unlink()
        # identify 0B or absent indices QL
        for f in self._paths["indices"]:
            #~ print(f, self._paths["indices"][f])
            for p in self._paths["indices"][f].glob("*_MSIL2A_*/"):
                if p.is_file():
                    logger.info("Identified old indice format {}".format(p.name))
                    if remove:
                        logger.info("Removing old indice format {}".format(p.name))
                        p.unlink()
                else:
                    for q in p.glob("*_QUICKLOOK.tif"):
                        if not ((q.stat().st_size == 3617212) or 
                                (q.stat().st_size == 4196652) or
                                (q.stat().st_size == 3617478)):
                            logger.info("Corrupted indice QL {} (bad size)".format(q.name))
                            if remove:
                                logger.info("Removing indice QL {}".format(q.name))
                                q.unlink()
    
                
        
        
        


        
