# -*- coding: utf-8 -*-

"""
This module contains usefull functions commons to every modules.
"""

import logging
from itertools import zip_longest
from datetime import datetime
from typing import Iterable


logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


# useful dates formats encountered when working with Sentinel-2 data
DATES_FORMATS = dict(
    ymd="%Y-%m-%d",
    filename="%Y%m%dT%H%M%S",
    metadata="%Y-%m-%dT%H:%M:%S.%fZ")


def format_word(word: str) -> str:
    """
    Formats a word to lower case and removes heading and trailing spaces.

    :param word: a string to format.
    """
    if not isinstance(word, str):
        raise TypeError("word has to be a string")
    return word.lower().strip()


def grouper(iterable, n: int, fillvalue: bool = None) -> Iterable:
    """
    Collects data into fixed-length chunks or blocks.

    :param iterable: an iterable.
    :param n: number of items per chunk/block.
    :param fillvalue: replacing value for missing items in the chunk.
    """
    args = [iter(iterable)] * n
    return zip_longest(*args, fillvalue=fillvalue)


def str_to_datetime(date_string: str, date_format: str) -> datetime:
    """
    Returns a datetime object from a string according to a format.

    :param date_string: string to format.
    :param date_format: date format.
    """
    try:
        date_format = DATES_FORMATS.get(date_format)
    except KeyError:
        raise ValueError("Invalid date format")
    return datetime.strptime(date_string, date_format)


def datetime_to_str(date: datetime, date_format: str) -> str:
    """
    Returns a string from a datetime object according to a format.

    :param date: string to format.
    :param date_format: date format.
    """
    try:
        date_format = DATES_FORMATS.get(date_format)
    except KeyError:
        raise ValueError("Invalid date format")
    return datetime.strftime(date, date_format)
